/* Copyright 2021 Jesse Haber-Kucharsky */
/* SPDX-License-Identifier: GPL-3.0-only */

#pragma once

#include <functional>
#include <memory>
#include <type_traits>

template <typename T>
struct List final {
  List() = default;
  List(T const &data, List const &l) : _head{std::make_shared<Node const>(data, l._head)} {
  }
  template <typename... Ts>
  explicit List(Ts const &...ts) : _head{(ts + ... + List{})._head} {
  }

  bool is_empty() const noexcept {
    return _head == nullptr;
  }

  T const &head() const noexcept {
    return _head->data;
  }

  List tail() const {
    return List{_head->next};
  }

 private:
  struct Node final {
    T data;
    std::shared_ptr<Node const> next;

    Node(T data, std::shared_ptr<Node const> next) : data{std::move(data)}, next{std::move(next)} {
    }
  };

  std::shared_ptr<Node const> _head{nullptr};

  explicit List(std::shared_ptr<Node const> ptr) : _head{std::move(ptr)} {
  }

 public:
  class const_iterator final {
   public:
    using difference_type = std::ptrdiff_t;
    using value_type = T const;
    using pointer = T const *;
    using reference = T const &;
    using iterator_category = std::forward_iterator_tag;

    const_iterator() : _node{nullptr} {
    }
    explicit const_iterator(Node const *ptr) : _node(ptr) {
    }

    reference operator*() const {
      return _node->data;
    }
    pointer operator->() {
      return &(_node->data);
    }

    const_iterator &operator++() {
      _node = _node->next.get();
      return *this;
    }
    const_iterator operator++(int) {
      const_iterator tmp = *this;
      ++(*this);
      return tmp;
    }

    friend bool operator==(const_iterator const &lhs, const_iterator const &rhs) {
      return lhs._node == rhs._node;
    }
    friend bool operator!=(const_iterator const &lhs, const_iterator const &rhs) {
      return lhs._node != rhs._node;
    }

   private:
    Node const *_node;
  };

  const_iterator cbegin() const {
    return const_iterator{_head.get()};
  }
  const_iterator cend() const {
    return const_iterator{nullptr};
  };
};

template <typename T>
List<T> operator+(T const x, List<T> const &xs) {
  return List<T>{x, xs};
}

template <typename T>
List<T> operator+(List<T> const &lhs, List<T> const &rhs) {
  if (lhs.is_empty()) {
    return rhs;
  }

  return lhs.head() + (lhs.tail() + rhs);
}

template <typename R, typename T, typename F>
R fold_right(F f, List<T> const &xs, R const &z) {
  if (xs.is_empty()) {
    return z;
  }

  return f(xs.head(), fold_right(f, xs.tail(), z));
}

template <typename T, typename F, typename R = std::invoke_result_t<F, T>>
auto map(F f, List<T> const &xs) {
  if (xs.is_empty()) {
    return List<R>{};
  }

  return f(xs.head()) + map(std::move(f), xs.tail());
}

template <typename T>
List<T> flatten(List<List<T>> const &xss) {
  return fold_right([](List<T> const &xs, List<T> const &ys) { return xs + ys; }, xss, List<T>{});
}
