/* Copyright 2021 Jesse Haber-Kucharsky */
/* SPDX-License-Identifier: GPL-3.0-only */

#include "list.hpp"

#include <pretty.hpp>

#include <charconv>
#include <cstring>
#include <iostream>
#include <memory>
#include <string>
#include <utility>
#include <variant>
#include <vector>

class Xml;

using Pair = std::pair<std::string, std::string>;
using Attributes = List<Pair>;

struct Element final {
  std::string name;
  Attributes attrs;
  List<std::shared_ptr<Xml const>> xmls;
};

class Xml final {
 public:
  explicit Xml(std::string s) : _data{std::move(s)} {
  }
  explicit Xml(Element e) : _data{std::move(e)} {
  }

  pretty::Doc prettify() const {
    List<pretty::Doc> const ds = PrettifyImpl::pp_xmls(*this);
    return pretty::fold(ds.cbegin(), ds.cend(), pretty::concat);
  }

 private:
  std::variant<std::string, Element> _data;

  struct PrettifyImpl {
    static List<pretty::Doc> words(std::string const& s) {
      std::vector<pretty::Doc> tokens{};
      pretty::tokenize(s, [&tokens](pretty::Doc const& d) { tokens.push_back(d); });
      List<pretty::Doc> ds{};
      auto iter = tokens.rbegin();

      while (iter != tokens.rend()) {
        ds = *iter++ + ds;
      }

      return ds;
    }

    static std::string quoted(std::string const& s) {
      return '"' + s + '"';
    }

    static List<pretty::Doc> pp_attr(Pair const& attr) {
      return List<pretty::Doc>{pretty::text(attr.first) + pretty::char_('=') +
                               pretty::text(quoted(attr.second))};
    }

    template <typename T, typename F>
    static pretty::Doc pp_fill(F f, List<T> const& ds) {
      if (ds.is_empty()) {
        return pretty::nil();
      }

      auto const ys = flatten(map(std::move(f), ds));
      return pretty::bracket("", pretty::fill(ys.cbegin(), ys.cend()), "");
    }

    static pretty::Doc pp_tag(std::string const& name, Attributes const& attrs) {
      return pretty::text(name) + pp_fill(pp_attr, attrs);
    }

    static List<pretty::Doc> pp_xmls(Xml const& x) {
      if (std::string const* text = std::get_if<std::string>(&(x._data)); text) {
        return words(*text);
      }

      Element const& e = std::get<Element>(x._data);

      if (e.xmls.is_empty()) {
        return List<pretty::Doc>{pretty::char_('<') + pp_tag(e.name, e.attrs) + pretty::text("/>")};
      } else {
        return List<pretty::Doc>{
            pretty::char_('<') + pp_tag(e.name, e.attrs) + pretty::char_('>') +
            pp_fill(pp_xmls, map([](std::shared_ptr<Xml const> const& p) { return *p; }, e.xmls)) +
            pretty::text("</") + pretty::text(e.name) + pretty::char_('>')};
      }
    }
  };
};

template <typename... Xs>
List<std::shared_ptr<Xml const>> with(Xs const&... xs) {
  return List<std::shared_ptr<Xml const>>{std::make_shared<Xml const>(xs)...};
}

static std::size_t parse_width(int argc, char** argv) {
  if (argc > 2) {
    std::cerr << "Usage: " << argv[0] << " [WIDTH]\n";
    std::exit(EXIT_FAILURE);
  } else if (argc == 1) {
    return 60;
  }

  std::size_t width{0};
  char const* const begin = argv[1];
  char const* const end = begin + std::strlen(begin);
  auto const result = std::from_chars(begin, end, width, 10);

  if ((result.ptr == end) && (result.ec == std::errc{})) {
    return width;
  }

  std::cerr << "Invalid width: '" << argv[1] << "'\n";
  std::exit(EXIT_FAILURE);
}

int main(int argc, char** argv) {
  Xml const x{
      Element{"p",
              Attributes{Pair{"color", "red"}, Pair{"font", "Times"}, Pair{"size", "10"}},
              with("Here is some",
                   Element{"em", Attributes{}, with("emphasized")},
                   "text. Here is a",
                   Element{"a", Attributes{Pair{"href", "http://example.com"}}, with("link")},
                   "elsewhere.")}};

  std::cout << x.prettify().fit_to(parse_width(argc, argv)) << std::endl;
}
