/* Copyright 2021 Jesse Haber-Kucharsky */
/* SPDX-License-Identifier: GPL-3.0-only */

#pragma once

#include <algorithm>
#include <concepts>
#include <cstring>
#include <forward_list>
#include <functional>
#include <iosfwd>
#include <iterator>
#include <memory>
#include <string>
#include <utility>
#include <vector>

namespace pretty {

namespace detail {

class DocImpl;

class LayoutImpl;

template <typename...>
struct FillImpl;
}  // namespace detail

class Visitor {
 public:
  Visitor() = default;
  Visitor(Visitor const &) = delete;
  Visitor(Visitor &&) = delete;

  Visitor &operator=(Visitor const &) = delete;
  Visitor &operator=(Visitor &&) = delete;

  virtual ~Visitor() = default;

  virtual void nil() = 0;

  virtual void char_(char) = 0;

  virtual void text(std::string const &) = 0;

  virtual void line(std::size_t) = 0;
};

template <typename F>
class LineVisitor final : public Visitor {
 public:
  LineVisitor(F f) : _f{std::move(f)} {
  }

  void nil() override {
    _f(_current);
  }

  void char_(char ch) override {
    _current += ch;
  }

  void text(std::string const &s) override {
    _current += s;
  }

  void line(std::size_t indent) override {
    _f(_current);
    _current = std::string(indent, ' ');
  }

 private:
  std::string _current{};
  F _f;
};

class Layout final {
 public:
  std::string to_string() const;

  void visit(Visitor &) const;

 private:
  std::shared_ptr<detail::LayoutImpl const> _impl;

  explicit Layout(std::shared_ptr<detail::LayoutImpl const>);

  friend class detail::LayoutImpl;
  friend class Doc;

  friend std::ostream &operator<<(std::ostream &, Layout const &);
};

std::ostream &operator<<(std::ostream &, Layout const &);

template <typename F>
void each_line(Layout const &l, F &&f) {
  auto v = LineVisitor{std::forward<F>(f)};
  l.visit(v);
}

class Doc;

Doc const &nil();
Doc const &line();
Doc const &cut();
Doc const &space();
Doc char_(char);
Doc text(std::string const &);
Doc concat(Doc const &, Doc const &);
Doc nest(std::size_t indent, Doc const &);
Doc group(Doc const &);
Doc beside(Doc const &left, Doc const &right);
Doc above(Doc const &top, Doc const &bottom);
Doc bracket(std::string const &left, Doc const &body, std::string const &right);

class Doc final {
 public:
  struct Nil final {};
  struct Line final {};
  struct Char final {};
  struct Text final {};
  struct Concat final {};
  struct Nest final {};
  struct Group final {};

  Doc();
  explicit Doc(Nil);
  explicit Doc(Line);
  Doc(Char, char);
  Doc(Text, std::string const &);
  Doc(Concat, Doc const &, Doc const &);
  Doc(Nest, std::size_t, Doc const &);
  Doc(Group, Doc const &);

  Doc operator+(Doc const &) const;

  Layout fit_to(std::size_t width) const;

 private:
  struct Union final {};

  std::shared_ptr<detail::DocImpl const> _impl;

  explicit Doc(std::shared_ptr<detail::DocImpl const>);
  Doc(Union, std::function<Doc()>, Doc const &);

  Doc flatten() const;

  friend class detail::DocImpl;
  friend class Layout;

  template <typename...>
  friend struct detail::FillImpl;

  template <std::input_iterator I>
  friend Doc fill(I, I);
};

template <typename F>
void tokenize(std::string const &str, F f) {
  std::vector<char> data{};
  std::copy(str.cbegin(), str.cend(), std::back_inserter(data));
  data.push_back('\0');
  char *token = std::strtok(data.data(), " \t\r\n");

  while (token != nullptr) {
    f(text(token));
    token = std::strtok(nullptr, " \t\r\n");
  }
}

namespace detail {

template <typename F, typename... Ds>
struct FoldImpl;

template <typename F>
struct FoldImpl<F> {
  static Doc apply(F) {
    return nil();
  }
};

template <typename F, typename D>
struct FoldImpl<F, D> {
  static Doc apply(F, D const &d) {
    return d;
  }
};

template <typename F, typename D, typename... Ds>
struct FoldImpl<F, D, Ds...> {
  static Doc apply(F f, D const &d, Ds const &...ds) {
    return f(d, FoldImpl<F, Ds...>::apply(f, ds...));
  }
};

template <>
struct FillImpl<> {
  static Doc apply() {
    return nil();
  }
};

template <typename D>
struct FillImpl<D> {
  static Doc apply(D const &d) {
    return d;
  }
};

template <typename D1, typename D2, typename... Ds>
struct FillImpl<D1, D2, Ds...> {
  static Doc apply(D1 const &d1, D2 const &d2, Ds const &...ds) {
    auto left = [d1, d2, ds...] { return d1.flatten() + cut() + fill(d2.flatten(), ds...); };
    Doc right = above(d1, fill(d2, ds...));
    return Doc{Doc::Union{}, std::move(left), std::move(right)};
  }
};

}  // namespace detail

template <typename F, std::same_as<Doc>... Ds>
Doc fold(F f, Ds const &...ds) {
  return detail::FoldImpl<F, Ds...>::apply(std::move(f), ds...);
}
template <std::input_iterator I, typename F>
Doc fold(I begin, I end, F f) {
  std::vector<Doc> ds(begin, end);

  if (ds.empty()) {
    return nil();
  }

  if (ds.size() == 1) {
    return ds.front();
  }

  Doc result = ds.back();
  auto iter = std::next(ds.rbegin());

  while (iter != ds.rend()) {
    result = f(*iter, result);
    iter = std::next(iter);
  }

  return result;
}

template <std::same_as<Doc>... Ds>
Doc spread(Ds const &...ds) {
  return fold(beside, ds...);
}
template <std::input_iterator I>
Doc spread(I begin, I end) {
  return fold(begin, end, beside);
}

template <std::same_as<Doc>... Ds>
Doc stack(Ds const &...ds) {
  return fold(above, ds...);
}
template <std::input_iterator I>
Doc stack(I begin, I end) {
  return fold(begin, end, above);
}

template <std::same_as<Doc>... Ds>
Doc fill(Ds const &...ds) {
  return detail::FillImpl<Ds...>::apply(ds...);
}
template <std::input_iterator I>
Doc fill(I begin, I end) {
  if (begin == end) {
    return nil();
  }

  Doc x = *begin++;

  if (begin == end) {
    return x;
  }

  Doc y = *begin++;
  std::forward_list<Doc> ds(begin, end);

  auto left = [x, y, ds]() mutable {
    ds.push_front(y.flatten());
    return beside(x.flatten(), fill(ds.begin(), ds.end()));
  };

  ds.push_front(y);
  Doc const right = above(x, fill(ds.begin(), ds.end()));
  return Doc{Doc::Union{}, std::move(left), right};
}

Doc fill_words(std::string const &);

}  // namespace pretty
