/* Copyright 2021 Jesse Haber-Kucharsky */
/* SPDX-License-Identifier: GPL-3.0-only */

#include "pretty.hpp"

#include <functional>
#include <iostream>

namespace pretty {

template <typename T>
using Ptr = std::shared_ptr<T const>;

template <typename T, typename... Args>
Ptr<T> make(Args &&...args) {
  return std::make_shared<T const>(std::forward<Args>(args)...);
}

template <typename T>
class LazyPtr final {
 public:
  template <typename F>
  explicit LazyPtr(F f) : _f{std::move(f)} {
  }

  Ptr<T> const &force() const {
    // The function is checked instead of the pointer because the result of the lazy expression
    // could be `nullptr`.
    if (_f != nullptr) {
      _ptr = _f();
      _f = nullptr;
    }

    return _ptr;
  }

 private:
  mutable std::function<Ptr<T>()> _f;
  mutable Ptr<T> _ptr{nullptr};
};

struct StringBuilder final : public Visitor {
  std::string value{};

  void nil() override {
    // Nothing.
  }

  void char_(char ch) override {
    value += ch;
  }

  void text(std::string const &s) override {
    value += s;
  }

  void line(std::size_t indent) override {
    value += '\n' + std::string(indent, ' ');
  }
};

struct StreamWriter final : public Visitor {
  std::ostream &os;

  explicit StreamWriter(std::ostream &os) : os{os} {
  }

  void nil() override {
    // Nothing.
  }

  void char_(char ch) override {
    os << ch;
  }

  void text(std::string const &s) override {
    os << s;
  }

  void line(std::size_t indent) override {
    os << '\n';

    for (std::size_t i = 0; i < indent; ++i) {
      os << ' ';
    }
  }
};

namespace detail {

class LayoutImpl;

using NextPtr = LazyPtr<LayoutImpl>;

class LayoutImpl {
 public:
  LayoutImpl() = default;
  LayoutImpl(LayoutImpl const &) = delete;
  LayoutImpl(LayoutImpl &&) = delete;

  LayoutImpl &operator=(LayoutImpl const &) = delete;
  LayoutImpl &operator=(LayoutImpl &&) = delete;

  virtual ~LayoutImpl() = default;

  virtual bool fits(std::size_t width, std::size_t offset) const = 0;

  virtual void visit(Visitor &) const = 0;
};

static bool fits_impl(LayoutImpl const *impl, std::size_t width, std::size_t offset) {
  if (offset > width) {
    return false;
  }

  if (impl == nullptr) {
    return true;
  }

  return impl->fits(width, offset);
}

static void visit_impl(LayoutImpl const *impl, Visitor &v) {
  if (impl == nullptr) {
    v.nil();
  } else {
    impl->visit(v);
  }
}

class CharLayout final : public LayoutImpl {
 public:
  template <typename Next>
  CharLayout(char value, Next next) : _value{value}, _next{std::move(next)} {
  }

  bool fits(std::size_t width, std::size_t offset) const override {
    return width - offset >= 1 && fits_impl(_next.force().get(), width, offset + 1);
  }

  void visit(Visitor &v) const override {
    v.char_(_value);
    visit_impl(_next.force().get(), v);
  }

 private:
  char _value;
  NextPtr _next;
};

class TextLayout final : public LayoutImpl {
 public:
  template <typename Next>
  TextLayout(Ptr<std::string> value, Next next) : _value{std::move(value)}, _next{std::move(next)} {
  }

  bool fits(std::size_t width, std::size_t offset) const override {
    std::size_t const size = _value->size();
    return size <= width - offset && fits_impl(_next.force().get(), width, offset + size);
  }

  void visit(Visitor &v) const override {
    v.text(*_value);
    visit_impl(_next.force().get(), v);
  }

 private:
  Ptr<std::string> _value;
  NextPtr _next;
};

class LineLayout final : public LayoutImpl {
 public:
  template <typename Next>
  LineLayout(std::size_t indent, Next next) : _indent{indent}, _next{std::move(next)} {
  }

  bool fits(std::size_t, std::size_t) const override {
    return true;
  }

  void visit(Visitor &v) const override {
    v.line(_indent);
    visit_impl(_next.force().get(), v);
  }

 private:
  std::size_t _indent;
  NextPtr _next;
};

}  // namespace detail

Layout::Layout(std::shared_ptr<detail::LayoutImpl const> impl) : _impl{std::move(impl)} {
}

void Layout::visit(Visitor &v) const {
  visit_impl(_impl.get(), v);
}

std::string Layout::to_string() const {
  StringBuilder builder{};
  this->visit(builder);
  return builder.value;
}

std::ostream &operator<<(std::ostream &os, Layout const &l) {
  StreamWriter writer{os};
  l.visit(writer);
  return os;
}

namespace detail {

class WorkList final {
 public:
  WorkList() = default;

  void push(std::size_t i, Ptr<DocImpl> x) {
    _head = make<Node>(i, std::move(x), _head);
  }

  bool is_empty() const {
    return _head == nullptr;
  }

  std::pair<std::size_t, Ptr<DocImpl>> pop() {
    auto result = std::pair{_head->i, _head->x};
    _head = _head->next;
    return result;
  }

 private:
  struct Node final {
    std::size_t i;
    Ptr<DocImpl> x;
    Ptr<Node> next;

    Node(std::size_t i, Ptr<DocImpl> x, Ptr<Node> next)
        : i{i}, x{std::move(x)}, next{std::move(next)} {
    }
  };

  Ptr<Node> _head{nullptr};
};

class DocImpl : public std::enable_shared_from_this<DocImpl> {
 public:
  DocImpl() = default;
  DocImpl(DocImpl const &) = delete;
  DocImpl(DocImpl &&) = delete;

  DocImpl &operator=(DocImpl const &) = delete;
  DocImpl &operator=(DocImpl &&) = delete;

  virtual ~DocImpl() = default;

  virtual Ptr<DocImpl> flatten() const = 0;

  virtual Ptr<LayoutImpl> best(std::size_t i, std::size_t w, std::size_t k, WorkList) const = 0;
};

static Ptr<DocImpl> flatten_impl(DocImpl const *impl) {
  if (impl == nullptr) {
    return nullptr;
  }

  return impl->flatten();
}

static Ptr<LayoutImpl> best_impl(std::size_t w, std::size_t k, WorkList l) {
  while (true) {
    if (l.is_empty()) {
      return nullptr;
    }

    auto const [i, x] = l.pop();

    if (x == nullptr) {
      continue;
    }

    return x->best(i, w, k, std::move(l));
  }
}

class CharDoc final : public DocImpl {
 public:
  explicit CharDoc(char value) : _value{value} {
  }

  Ptr<DocImpl> flatten() const override {
    return this->shared_from_this();
  }

  Ptr<LayoutImpl> best(std::size_t, std::size_t w, std::size_t k, WorkList l) const override {
    return make<CharLayout>(_value,
                            [w, k, l = std::move(l)] { return best_impl(w, k + 1, std::move(l)); });
  }

 private:
  char _value;
};

struct LineDoc final : DocImpl {
  Ptr<DocImpl> flatten() const override {
    static auto const instance = make<CharDoc>(' ');
    return instance;
  }

  Ptr<LayoutImpl> best(std::size_t i, std::size_t w, std::size_t, WorkList l) const override {
    return make<LineLayout>(i, [w, i, l = std::move(l)] { return best_impl(w, i, std::move(l)); });
  }
};

class TextDoc final : public DocImpl {
 public:
  explicit TextDoc(std::string s) : _value{make<std::string>(std::move(s))} {
  }

  Ptr<DocImpl> flatten() const override {
    return this->shared_from_this();
  }

  Ptr<LayoutImpl> best(std::size_t, std::size_t w, std::size_t k, WorkList l) const override {
    return make<TextLayout>(_value, [w, k, l = std::move(l), n = _value->size()] {
      return best_impl(w, k + n, std::move(l));
    });
  }

 private:
  Ptr<std::string> _value;
};

class ConcatDoc final : public DocImpl {
 public:
  ConcatDoc(Ptr<DocImpl> left, Ptr<DocImpl> right)
      : _left{std::move(left)}, _right{std::move(right)} {
  }

  Ptr<DocImpl> flatten() const override {
    return make<ConcatDoc>(flatten_impl(_left.get()), flatten_impl(_right.get()));
  }

  Ptr<LayoutImpl> best(std::size_t i, std::size_t w, std::size_t k, WorkList l) const override {
    l.push(i, _right);
    l.push(i, _left);
    return best_impl(w, k, std::move(l));
  }

 private:
  Ptr<DocImpl> _left;
  Ptr<DocImpl> _right;
};

class NestDoc final : public DocImpl {
 public:
  explicit NestDoc(std::size_t indent, Ptr<DocImpl> body)
      : _indent{indent}, _body{std::move(body)} {
  }

  Ptr<DocImpl> flatten() const override {
    return make<NestDoc>(_indent, flatten_impl(_body.get()));
  }

  Ptr<LayoutImpl> best(std::size_t i, std::size_t w, std::size_t k, WorkList l) const override {
    l.push(i + _indent, _body);
    return best_impl(w, k, std::move(l));
  }

 private:
  std::size_t _indent;
  Ptr<DocImpl> _body;
};

class UnionDoc final : public DocImpl {
 public:
  template <typename F>
  UnionDoc(F left, Ptr<DocImpl> right) : _left{std::move(left)}, _right{std::move(right)} {
  }

  Ptr<DocImpl> flatten() const override {
    return flatten_impl(_left.force().get());
  }

  Ptr<LayoutImpl> best(std::size_t i, std::size_t w, std::size_t k, WorkList l) const override {
    WorkList ll = l;
    l.push(i, _left.force());
    auto longer = best_impl(w, k, std::move(l));

    if (fits_impl(longer.get(), w, k)) {
      return longer;
    }

    ll.push(i, _right);
    return best_impl(w, k, std::move(ll));
  }

 private:
  LazyPtr<DocImpl> _left;
  Ptr<DocImpl> _right;
};

}  // namespace detail

Doc::Doc() : _impl{nullptr} {
}
Doc::Doc(Nil) : Doc{} {
}
Doc::Doc(Line) : _impl{make<detail::LineDoc>()} {
}
Doc::Doc(Char, char ch) : _impl{make<detail::CharDoc>(ch)} {
}
Doc::Doc(Text, std::string const &s) : _impl{make<detail::TextDoc>(s)} {
}
Doc::Doc(Concat, Doc const &left, Doc const &right)
    : _impl{make<detail::ConcatDoc>(left._impl, right._impl)} {
}
Doc::Doc(Nest, std::size_t indent, Doc const &body)
    : _impl{make<detail::NestDoc>(indent, body._impl)} {
}
Doc::Doc(Group, Doc const &body)
    : _impl{make<detail::UnionDoc>([body = body._impl] { return body->flatten(); }, body._impl)} {
}
Doc::Doc(Union, std::function<Doc()> left, Doc const &right)
    : _impl{
          make<detail::UnionDoc>([left = std::move(left)] { return left()._impl; }, right._impl)} {
}
Doc::Doc(std::shared_ptr<detail::DocImpl const> ptr) : _impl{std::move(ptr)} {
}

Doc Doc::operator+(const Doc &other) const {
  return Doc{Concat{}, *this, other};
}

Layout Doc::fit_to(std::size_t width) const {
  detail::WorkList l{};
  l.push(0, _impl);
  return Layout{detail::best_impl(width, 0, std::move(l))};
}

Doc Doc::flatten() const {
  return Doc{flatten_impl(_impl.get())};
}

Doc const &nil() {
  static Doc const instance{Doc::Nil{}};
  return instance;
}

Doc const &line() {
  static Doc const instance{Doc::Line{}};
  return instance;
}

Doc const &cut() {
  static Doc const instance{Doc::Group{}, line()};
  return instance;
}

Doc const &space() {
  static Doc const instance{Doc::Char{}, ' '};
  return instance;
}

Doc char_(char ch) {
  return Doc{Doc::Char{}, ch};
}

Doc text(std::string const &s) {
  return Doc{Doc::Text{}, s};
}

Doc concat(Doc const &left, Doc const &right) {
  return Doc{Doc::Concat{}, left, right};
}

Doc nest(std::size_t indent, Doc const &body) {
  return Doc{Doc::Nest{}, indent, body};
}

Doc group(Doc const &body) {
  return Doc{Doc::Group{}, body};
}

Doc beside(Doc const &left, Doc const &right) {
  return left + space() + right;
}

Doc above(Doc const &left, Doc const &right) {
  return left + line() + right;
}

Doc bracket(std::string const &left, Doc const &body, std::string const &right) {
  return group(text(left) + nest(2, line() + body) + line() + text(right));
}

Doc fill_words(std::string const &str) {
  std::vector<Doc> ds{};
  tokenize(str, [&ds](Doc const &d) { ds.push_back(d); });

  return fold(std::make_move_iterator(ds.begin()),
              std::make_move_iterator(ds.end()),
              [](Doc const &left, Doc const &right) { return left + cut() + right; });
}

}  // namespace pretty
