/* Copyright 2021 Jesse Haber-Kucharsky */
/* SPDX-License-Identifier: GPL-3.0-only */

#include "pretty.hpp"

#include <string>
#include <vector>

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#define S " "
#define L "\n"

using namespace pretty;
using namespace testing;

static std::vector<std::string> lines_of(Doc const& d, std::size_t width = 10) {
  std::vector<std::string> lines{};
  each_line(d.fit_to(width), [&lines](std::string const& line) { lines.push_back(line); });
  return lines;
}

TEST(PrettyTest, Text) {
  EXPECT_THAT(lines_of(text("AAA")), ElementsAre("AAA"));
  // Exceeds the width.
  EXPECT_THAT(lines_of(text("AAAAAAAAAAAA")), ElementsAre("AAAAAAAAAAAA"));
}

TEST(PrettyTest, Concat) {
  EXPECT_THAT(lines_of(text("AAA") + text("BBB") + text("CCC")), ElementsAre("AAABBBCCC"));
}

TEST(PrettyTest, Char) {
  EXPECT_THAT(lines_of(char_('A') + char_('B') + char_('C')), ElementsAre("ABC"));
}

TEST(PrettyTest, Line) {
  EXPECT_THAT(lines_of(text("AAA") + line() + text("BBB") + line() + text("CCC")),
              ElementsAre("AAA", "BBB", "CCC"));
}

TEST(PrettyTest, Nest) {
  EXPECT_THAT(lines_of(nest(2, text("AAA") + line() + text("BBB") + line() + text("CCC"))),
              ElementsAre("AAA", S S "BBB", S S "CCC"));
}

TEST(PrettyTest, Group) {
  Doc const d = group(text("AAA") + line() + text("BBB"));
  EXPECT_THAT(lines_of(d, 5), ElementsAre("AAA", "BBB"));
  EXPECT_THAT(lines_of(d, 10), ElementsAre("AAA" S "BBB"));
}

TEST(PrettyTest, Beside) {
  EXPECT_THAT(lines_of(beside(text("AAA"), text("BBB"))), ElementsAre("AAA" S "BBB"));
}

TEST(PrettyTest, Above) {
  EXPECT_THAT(lines_of(above(text("AAA"), text("BBB"))), ElementsAre("AAA", "BBB"));
}

TEST(PrettyTest, Bracket) {
  EXPECT_THAT(lines_of(bracket("{", above(text("AAA"), text("BBB")), "}")),
              ElementsAre("{", S S "AAA", S S "BBB", "}"));
}

TEST(PrettyTest, Fold) {
  EXPECT_THAT(lines_of(fold(beside)), ElementsAre(""));
  EXPECT_THAT(lines_of(fold(beside, text("AAA"))), ElementsAre("AAA"));
  EXPECT_THAT(lines_of(fold(beside, text("AAA"), text("BBB"))), ElementsAre("AAA" S "BBB"));
}

TEST(PrettyTest, Fold_iter) {
  std::vector<Doc> ds1{};
  EXPECT_THAT(lines_of(fold(ds1.cbegin(), ds1.cend(), beside)), ElementsAre(""));
  auto ds2 = std::vector{text("AAA")};
  EXPECT_THAT(lines_of(fold(ds2.cbegin(), ds2.cend(), beside)), ElementsAre("AAA"));
  auto ds3 = std::vector{text("AAA"), text("BBB"), text("CCC")};
  EXPECT_THAT(lines_of(fold(ds3.cbegin(), ds3.cend(), beside)), ElementsAre("AAA" S "BBB" S "CCC"));
}

TEST(PrettyTest, Spread) {
  EXPECT_THAT(lines_of(spread(text("AAA"), char_('X'), text("BBB"), text("CCC"))),
              ElementsAre("AAA" S "X" S "BBB" S "CCC"));
}

TEST(PrettyTest, Spread_iter) {
  auto ds = std::vector{text("AAA"), char_('X'), text("BBB"), text("CCC")};
  EXPECT_THAT(lines_of(spread(ds.begin(), ds.end())), ElementsAre("AAA" S "X" S "BBB" S "CCC"));
}

TEST(PrettyTest, Stack) {
  EXPECT_THAT(lines_of(stack(text("AAA"), char_('X'), text("BBB"), text("CCC"))),
              ElementsAre("AAA", "X", "BBB", "CCC"));
}

TEST(PrettyTest, Stack_iter) {
  auto ds = std::vector{text("AAA"), char_('X'), text("BBB"), text("CCC")};
  EXPECT_THAT(lines_of(stack(ds.begin(), ds.end())), ElementsAre("AAA", "X", "BBB", "CCC"));
}

TEST(PrettyTest, cut) {
  Doc const d = text("AAA") + cut() + text("BBB") + cut() + text("CCC") + cut() + text("DDD");
  EXPECT_THAT(lines_of(d), ElementsAre("AAA" S "BBB", "CCC" S "DDD"));
}

TEST(PrettyTest, Doc_move) {
  auto d1 = stack(text("AAA"), text("BBB"), text("CCC"));
  auto d2 = std::move(d1);
  EXPECT_THAT(lines_of(d1), ElementsAre(""));
  EXPECT_THAT(lines_of(d2), ElementsAre("AAA", "BBB", "CCC"));
}

TEST(PrettyTest, Layout_move) {
  Doc const d = stack(text("AAA"), text("BBB"));
  Layout l1 = d.fit_to(20);
  Layout const l2 = std::move(l1);
  EXPECT_EQ(l1.to_string(), "");
  EXPECT_EQ(l2.to_string(), "AAA" L "BBB");
}
